"""
Bitboard operations. All operations assume that the board is 8x8.
"""
from typing import List
from collections import namedtuple
import operator as op
from functools import reduce

N_EDGE = 0xff00000000000000
E_EDGE = 0x0101010101010101
S_EDGE = 0x00000000000000ff
W_EDGE = 0x8080808080808080
POS_SLOPE = 0x0102040810204080
NEG_SLOPE = 0x8040201008040201
ALL_ = 0xffffffffffffffff
NONE = 0x0000000000000000
DIRECTIONS = frozenset({'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw'})

Position = namedtuple('Position', 'row, col', module=__name__)

def pos_mask(row: int, col: int) -> int:
    """ Return the bit mask for (row, col) on a bitboard. """
    assert 0 <= row < 8
    assert 0 <= col < 8
    return 0x8000000000000000 >> col >> row * 8

def shift(bits: int, dir_: str, times: int = 1) -> int:
    """ Shift `bits` in direction `dir_` `times` times. """
    dir_ = dir_.lower()
    res = bits
    if 'n' in dir_:
        res <<= 8 * times
    if 'e' in dir_:
        # wall is a bit mask that ensures bits aren't shifted beyond where they are meant to go
        wall = not_(reduce(op.or_, (W_EDGE >> shift_ for shift_ in range(times)), 0))
        res >>= 1 * times
        res &= wall
    if 's' in dir_:
        res >>= 8 * times
    if 'w' in dir_:
        # wall is a bit mask that ensures bits aren't shifted beyond where they are meant to go
        wall = not_(reduce(op.or_, (E_EDGE << shift_ for shift_ in range(times)), 0))
        res <<= 1 * times
        res &= wall
    return res & ALL_

def on_edge(bits: int) -> str:
    """ Returns the edges that contain on-bits. """
    edges = []
    if bits & N_EDGE:
        edges.append('n')
    if bits & E_EDGE:
        edges.append('e')
    if bits & S_EDGE:
        edges.append('s')
    if bits & W_EDGE:
        edges.append('w')
    return ''.join(edges)

def dilate(bits: int, dir_: str, times: int = 1) -> int:
    for _ in range(times):
        bits |= shift(bits, dir_)
    return bits

def not_(bits: int) -> int:
    return ~bits & ALL_

def to_list(bits: int) -> List[Position]:
    positions = []
    for r in range(8):
        for c in range(8):
            mask = pos_mask(r, c)
            if bits & mask > 0:
                positions.append(Position(r, c))
    return positions



from __future__ import annotations
import tkinter as tk
import tkinter.messagebox  # pylint: disable=unused-import
import logging
from queue import Queue, Empty as QueueEmpty
from random import choice as chooseFrom
from . import bitboard as bb
from .game import Game, GameType, EventName, BoardState
from .player import Color


EVENT_QUEUE = Queue()
_EVENT_SUBSCRIPTIONS = dict()
_LOGGER = logging.getLogger(__name__)
_game = None  # pylint: disable=invalid-name


class BoardView(tk.Canvas):
    def __init__(self, master):
        self._cell_size = 40
        self._border_width = 0
        self._grid_line_width = 1
        super().__init__(
            master,
            width=self._cell_size * 8 + self._grid_line_width * 7,
            height=self._cell_size * 8 + self._grid_line_width * 7,
            borderwidth=f'{self._border_width}',
            highlightthickness=0,
            bg='#000'
        )
        self.pack()
        self.bind('<Button-1>', self.onclick)

        self._board_state = None
        _LOGGER.debug('Canvas dim (%d, %d)',
                      self.winfo_reqwidth(),
                      self.winfo_reqheight())
        _subscribe(EventName.BOARD_CHANGED, self._redraw)
        self._redraw(0, 0)

    def onclick(self, event):
        _LOGGER.debug('BoardView clicked %s', event)
        if _game is None:
            return
        col = (event.x - self._border_width) / self.winfo_reqwidth() * 8
        row = (event.y - self._border_width) / self.winfo_reqheight() * 8
        _LOGGER.debug('(row=%f, col=%f)', row, col)
        _LOGGER.debug('(row=%d, col=%d)', row, col)
        if 0 <= col < 8 and 0 <= row < 8:
            move = bb.Position(int(row), int(col))
            if move in _game.board_state.valid_moves():
                _game.my_player.move = move
            else:
                _LOGGER.info('%s played an invalid move', _game.my_player)

    def _redraw(self, white: int, black: int):
        _LOGGER.debug('Redrawing Board with white=%0#16x, black=%0#16x', white, black)
        self.delete(tk.ALL)
        for r in range(8):
            for c in range(8):
                x_off = c * self._cell_size + c * self._grid_line_width + self._border_width
                y_off = r * self._cell_size + r * self._grid_line_width + self._border_width
                # Draw green empty cell
                self.create_rectangle(
                    x_off,
                    y_off,
                    self._cell_size + x_off,
                    self._cell_size + y_off,
                    fill='#060'
                )

                # Draw piece in cell
                mask = bb.pos_mask(r, c)
                inset = 5
                fill_color = None
                if white & mask != 0:
                    fill_color = '#fff'
                elif black & mask != 0:
                    fill_color = '#000'
                if fill_color:
                    self.create_oval(
                        x_off + inset,
                        y_off + inset,
                        x_off + self._cell_size - inset,
                        y_off + self._cell_size - inset,
                        fill=fill_color)


class NewGameDialog(tk.Toplevel):
    def __init__(self, parent, board_view: BoardView):
        super().__init__(parent)
        self.transient(parent)
        self.focus_set()
        self.grab_set()
        self.title('New Game')
        self.board_view = board_view

        frame = tk.Frame(self)
        frame.pack_configure(expand=True)
        self.game_type = tk.StringVar(frame, GameType.COMPUTER.name)
        self.color_var = tk.StringVar(frame, Color.BLACK.name)
        oppenent_frame = tk.LabelFrame(frame, text='Opponent')
        oppenent_frame.grid(row=0, column=0, sticky='n')
        tk.Radiobutton(
            oppenent_frame,
            text='Computer',
            variable=self.game_type,
            value=GameType.COMPUTER.name
        ).grid(row=1, column=0, sticky='w')
        tk.Radiobutton(
            oppenent_frame,
            text='Online',
            variable=self.game_type,
            value=GameType.ONLINE.name,
            state=tk.DISABLED
        ).grid(row=2, column=0, sticky='w')
        color_frame = tk.LabelFrame(frame, text='Color')
        color_frame.grid(row=0, column=1, sticky='n')
        tk.Radiobutton(
            color_frame,
            text='Black',
            variable=self.color_var,
            value=Color.BLACK.name
        ).grid(row=1, column=1, sticky='w')
        tk.Radiobutton(
            color_frame,
            text='White',
            variable=self.color_var,
            value=Color.WHITE.name
        ).grid(row=2, column=1, sticky='w')
        tk.Radiobutton(
            color_frame,
            text='Random',
            variable=self.color_var,
            value=None
        ).grid(row=3, column=1, sticky='w')
        tk.Button(frame, text='Cancel', command=self.cancel).grid(row=1, column=0)
        tk.Button(frame, text='Start', command=self.submit).grid(row=1, column=1)
        self.bind('<Escape>', lambda e: self.cancel())
        self.bind('<Return>', lambda e: self.submit())
        self.wait_window(self)
        _LOGGER.debug('NewGameDialog closed')

    def cancel(self):
        _LOGGER.debug('Cancel')
        self.destroy()

    def submit(self):
        _LOGGER.debug('Submit')
        global _game  # pylint: disable=invalid-name

        if _game is not None:
            proceed = tk.messagebox.askokcancel(
                title='Start New Game',
                message='Another game is already in progress. Proceed?'
            )
            if not proceed:
                self.destroy()

        state = BoardState(
            init_white=0x0000001008000000,
            init_black=0x0000000810000000,
            turn_player_color=Color.BLACK
        )
        self.board_view.board_state = state
        if _game:
            _game.interrupt()
        _game = Game(
            state,
            self.color_var.get() and Color[self.color_var.get()] or chooseFrom(list(Color)),
            GameType[self.game_type.get()],
            EVENT_QUEUE
        )
        _game.start()
        self.destroy()


def _subscribe(name: EventName, callback):
    if name in _EVENT_SUBSCRIPTIONS:
        _EVENT_SUBSCRIPTIONS[name].append(callback)
    else:
        _EVENT_SUBSCRIPTIONS[name] = [callback]


def _poll_queue(root: tk.Tk):
    try:
        event_name, *args = EVENT_QUEUE.get_nowait()
    except QueueEmpty:
        pass
    else:
        if event_name in _EVENT_SUBSCRIPTIONS:
            for callback in _EVENT_SUBSCRIPTIONS[event_name]:
                callback(*args)
        else:
            _LOGGER.warning('Unhandled event %r with args %s', event_name, args)
    root.after(100, _poll_queue, root)


def _init_widgets(root: tk.Tk):
    root.title('Othello')

    main = tk.Frame(root)
    main.pack_configure(expand=True)
    board_view = BoardView(main)

    menubar = tk.Menu(root)
    gamemenu = tk.Menu(menubar, tearoff=0)
    gamemenu.add_command(label='New Game...', command=lambda: NewGameDialog(root, board_view))
    menubar.add_cascade(label='Game', menu=gamemenu)
    root.config(menu=menubar)


def loop():
    _LOGGER.info('Init ui loop')
    try:
        root = tk.Tk()
        _init_widgets(root)
        root.after(100, _poll_queue, root)
        root.mainloop()
        global _game  # pylint: disable=invalid-name
        if _game is not None:
            _game.interrupt()
            _game.join()
    finally:
        _LOGGER.info('Quit ui loop')



import logging
import os
logging.basicConfig(style='{',
                    format='[{name}:{levelname}] - {threadName} - In {funcName}: {message}',
                    level=getattr(logging, os.getenv('LOG', ''), logging.WARNING))



from __future__ import annotations
from typing import TYPE_CHECKING, Callable
import asyncio
import enum
from . import ai

if TYPE_CHECKING:
    from othello.game import BoardState
    from othello.bitboard import Position


class Color(enum.Enum):
    """
    Represents color of pieces.
    There can only be two colors.
    """
    BLACK = enum.auto()
    WHITE = enum.auto()


class Player:
    def __init__(self, color: Color):
        self._color = color
        self._loop = asyncio.get_running_loop()
        self._move = self._loop.create_future()

    @property
    def color(self):
        return self._color

    @property
    async def move(self) -> Position:
        """ The next move this player intends to make. """
        move = await self._move
        del self.move
        return move

    @move.setter
    def move(self, value: Position):
        def callback():
            if self._move.done():
                raise RuntimeError(f'Attempt to set {self} move but it has already been set')
            self._move.set_result(value)
        self._loop.call_soon_threadsafe(callback)

    @move.deleter
    def move(self):
        self._move = self._loop.create_future()

    def __repr__(self):
        return f'Player({self._color})'


class AIPlayer(Player):
    def __init__(self, color: Color, state: BoardState, strat: Callable[[BoardState], Position] = ai.random):
        super().__init__(color)
        self._strat = strat
        self._state = state

    @Player.move.getter  # pylint: disable=no-member
    async def move(self):  # pylint: disable=invalid-overridden-method
        return self._strat(self._state)



import logging
from othelloai import ui

_LOGGER = logging.getLogger('othello')

try:
    ui.loop()  # Blocked until application is closed or an error occurs
except KeyboardInterrupt:
    _LOGGER.info('Quitting application via KeyboardInterrupt...')
finally:
    logging.shutdown()



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['othelloai']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'othello-ai',
    'version': '0.1.0',
    'description': 'Othello game engine',
    'long_description': '# othello\nAn othello game engine\n\n## Dev environment\nFor the dev environment I recommend using [pyenv](https://github.com/pyenv/pyenv#installation) and [poetry](https://poetry.eustace.io/docs/#installation)\n1. Download and install \n1. Run `poetry install` \n1. Run `poetry run python -m othello`',
    'author': 'Jason R. Carrete',
    'author_email': 'jasoncarrete5@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jcarrete5/othello',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations
from random import choice as chooseFrom
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from othello import bitboard as bb
    from othello.game import BoardState

def random(state: BoardState) -> bb.Position:
    """ Returns a random valid move or None if no moves are possible. """
    try:
        return chooseFrom(state.valid_moves())
    except IndexError:
        return None



from __future__ import annotations
import logging
import re
import threading
import asyncio
import enum
from typing import TYPE_CHECKING, List
from . import bitboard as bb
from .player import Color, Player, AIPlayer

if TYPE_CHECKING:
    from queue import Queue

_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.INFO)


class EventName(enum.Enum):
    BOARD_CHANGED = enum.auto()
    ERROR = enum.auto()


class GameType(enum.Enum):
    COMPUTER = enum.auto()
    ONLINE = enum.auto()


class IllegalMoveError(Exception):
    """ Raised when an illegal move is attempted. """


class BoardState:
    def __init__(self, init_white=0, init_black=0, turn_player_color=Color.BLACK):
        self.white = init_white
        self.black = init_black
        self.turn_player_color = turn_player_color

    def empty_cells(self) -> int:
        return bb.not_(self.white | self.black)

    def valid_moves(self) -> List[bb.Position]:
        """ Return a list of positions of valid moves for the turn player. """
        if self.turn_player_color is Color.BLACK:
            my_pieces = self.black
            foe_pieces = self.white
        else:
            my_pieces = self.white
            foe_pieces = self.black
        moves_bb = 0x0
        for dir_ in bb.DIRECTIONS:
            candidates = foe_pieces & bb.shift(my_pieces, dir_)
            while candidates != 0:
                shifted = bb.shift(candidates, dir_)
                moves_bb |= self.empty_cells() & shifted
                candidates = foe_pieces & shifted
        return bb.to_list(moves_bb)

    def reset(self):
        self.white = 0
        self.black = 0
        self.turn_player_color = Color.BLACK

    def __eq__(self, other: BoardState):
        return \
            self.white == other.white \
            and self.black == other.black \
            and self.turn_player_color == other.turn_player_color

    def __str__(self):
        def symbol_at(row: int, col: int) -> str:
            mask = bb.pos_mask(row, col)
            if self.white & mask != 0:
                return '\u2588'
            elif self.black & mask != 0:
                return '\u2591'
            else:
                return '-'
        res = [symbol_at(r, c) for r in range(8) for c in range(8)]
        return '\n'.join(re.findall('........', ''.join(res)))


class Board:
    def __init__(self, board_state: BoardState):
        self.board_state = board_state

    def place(self, color: Color, pos: bb.Position):
        """ Place a piece of color `color` at position `pos`

        Raises an IllegalMoveError if an illegal move was attempted
        """
        assert 0 <= pos.row < 8
        assert 0 <= pos.col < 8
        pos_mask = bb.pos_mask(*pos)
        self._capture(color, pos)
        if color is Color.BLACK:
            self.board_state.black |= pos_mask
        else:
            self.board_state.white |= pos_mask

    def _capture(self, color: Color, pos: bb.Position):
        """
        Find pieces that should be captured by playing `color`
        at `pos` and capture those pieces

        Raises an IllegalMoveError if an illegal move was attempted
        """
        my_pieces = self.board_state.white if color is Color.WHITE else self.board_state.black
        empty = self.board_state.empty_cells()
        class State:
            def __init__(self, dir_, init_bb=0x0):
                self.bb = init_bb
                self.dir = dir_
                self.capped = False
                self.on_edge = False
                self.on_empty = False

            def dilate(self):
                self.bb = bb.dilate(self.bb, self.dir)

            def should_commit(self):
                return self.capped

            def should_keep_dilating(self):
                selected = bb.not_(bb.pos_mask(*pos)) & self.bb
                self.on_edge = len(set(self.dir) & set(bb.on_edge(self.bb))) > 0
                self.capped = my_pieces & selected != 0
                self.on_empty = empty & selected != 0
                return not self.on_edge and not self.capped and not self.on_empty
        start = bb.pos_mask(*pos)
        states = [State(dir_, start) for dir_ in bb.DIRECTIONS]
        did_commit = False
        for state in states:
            while state.should_keep_dilating():
                state.dilate()
            if state.should_commit():
                did_commit = True
                if color is Color.WHITE:
                    self.board_state.white |= state.bb
                    self.board_state.black &= bb.not_(state.bb)
                elif color is Color.BLACK:
                    self.board_state.black |= state.bb
                    self.board_state.white &= bb.not_(state.bb)
        if not did_commit:
            raise IllegalMoveError

    def __str__(self):
        return str(self.board_state)


class Game(threading.Thread):
    game_counter = 0

    def __init__(self,
                 board_state: BoardState,
                 my_color: Color,
                 type_: GameType,
                 out_queue: Queue = None):
        super().__init__(name=f'GameThread ({Game.game_counter})')
        Game.game_counter += 1
        self.board_state = board_state
        self.my_player = None
        self._loop = None
        self._my_color = my_color
        self._board = Board(self.board_state)
        self._type = type_
        self._out_queue = out_queue

    def interrupt(self):
        if self._loop and self._loop.is_running():
            self._loop.call_soon_threadsafe(self._loop.stop)
        else:
            _LOGGER.warning("Game event loop was interrupted but it wasn't running")

    async def loop(self):
        self._loop = asyncio.get_running_loop()

        # Send initial board state
        if self._out_queue:
            self._out_queue.put(
                (EventName.BOARD_CHANGED, self.board_state.white, self.board_state.black)
            )

        self.my_player = Player(self._my_color)
        # Set opponent color to be opposite mine
        opp_color = list(Color)[0] if list(Color)[0] is not self.my_player.color else list(Color)[1]
        if self._type is GameType.COMPUTER:
            opponent = AIPlayer(opp_color, self.board_state)
        else:
            raise RuntimeError(f'{self._type} is not implemented')

        turn_player = self.my_player if self.my_player.color is Color.BLACK else opponent
        while self._loop.is_running():
            _LOGGER.info('Waiting for %s to make a move', turn_player)
            move = await turn_player.move
            if move is not None:
                _LOGGER.info('%s played %s', turn_player, move)
                self._board.place(turn_player.color, move)
            else:
                _LOGGER.info('%s passed their move', str(turn_player))

            if self._out_queue:
                self._out_queue.put(
                    (EventName.BOARD_CHANGED, self.board_state.white, self.board_state.black)
                )

            # Change to other player
            turn_player = self.my_player if turn_player is opponent else opponent
            self.board_state.turn_player_color = turn_player.color

    def run(self):
        _LOGGER.info('New game started')
        try:
            asyncio.run(self.loop())
        except RuntimeError as err:
            _LOGGER.error(err)
        finally:
            _LOGGER.info('Game ended')
